#include <iostream>
#include <fstream>
#include <vector>

#include <cassert>
#include <cstdint>

#pragma pack(push)
#pragma pack(2)

struct atari_header
{
	uint16_t magic;
	uint32_t textSize;
	uint32_t dataSize;
	uint32_t bssSize;
	uint32_t symbolTableSize;
	uint16_t reserved[5];
};

struct atari_symbol
{
	char name[8];
	uint16_t type;
	uint32_t value;
};

struct prg_section
{
	uint32_t start;
	uint32_t end;		// end like stl containers
	uint32_t pad;
};

#pragma pack(pop)

uint16_t big_endian(uint16_t v)
{
	return (v >> 8) | (v << 8);
}

uint32_t big_endian(uint32_t v)
{
	return big_endian((uint16_t)(v >> 16)) | (uint32_t)big_endian((uint16_t)(v)) << 16;
}

uint32_t relocate_atari(void *buffer, uint32_t org, std::vector<uint8_t> *output)
{
	atari_header *header = (atari_header *)buffer;
	
	if ( header->magic != big_endian((uint16_t)0x601a) )
	{
		std::cout << "Invalid TOS header" << std::endl;
		return 0;
	}
	
	uint8_t *text = (uint8_t *)buffer + sizeof(atari_header);
	uint8_t *relocTable = text + big_endian(header->textSize) + big_endian(header->dataSize) + big_endian(header->symbolTableSize);
	
	uint32_t offset = big_endian(*(uint32_t*)relocTable);
	relocTable += 4;
	
	uint32_t max = big_endian(header->textSize) + big_endian(header->dataSize);
	output->insert(output->begin(), text, text + max);
	
	if (offset)
	{
		uint32_t current = big_endian(*(uint32_t*)&text[offset]);
		*(uint32_t*)&output[offset] = big_endian(current + org);
		
		while ( *relocTable )
		{
			uint8_t lFix = *relocTable++;
			if( lFix == 1 )
				offset += 254;
			else
			{
				offset += lFix;
				if( offset < max )
				{
					current = big_endian(*(uint32_t*)&text[offset]);
					*(uint32_t*)&text[offset] = big_endian(current + org);
				}
			}
		}
	}

	// nuevo tamaño sin la tabla de simbolos ni la tabla de relocalizacion
	return big_endian(header->textSize) + big_endian(header->dataSize);
}

uint32_t shifted_address(uint32_t address, const prg_section *section)
{
	for ( int i = 0; i < 3; ++i, ++section )
	{
		if ( section->start <= address && address < section->end )
			return address + section->pad;
	}

	std::cerr << "Address not found in any zone: $" << std::hex << address << std::endl;
	return 0;
}

uint32_t relocate_jaguar(void *buffer, uint32_t org, std::vector<uint8_t> *output)
{
	atari_header *header = (atari_header *)buffer;
	
	if ( header->magic != big_endian((uint16_t)0x601a) )
	{
		std::cerr << "Invalid TOS header" << std::endl;
		return 0;
	}
	
	uint8_t *text = (uint8_t *)buffer + sizeof(atari_header);
	uint8_t *relocTable = text + big_endian(header->textSize) + big_endian(header->dataSize) + big_endian(header->symbolTableSize);
	
	uint32_t offset = big_endian(*(uint32_t*)relocTable);
	relocTable += 4;

	uint32_t max = big_endian(header->textSize) + big_endian(header->dataSize);

	prg_section sections[3];

	// insert text
	sections[0].start = 0;
	sections[0].end = (uint32_t)big_endian(header->textSize);
	sections[0].pad = 0;
	output->insert(output->end(), text, text + big_endian(header->textSize));

	// pad size
	sections[1].pad = sections[0].pad;
	while ( (output->size() & 7) != 0 )
	{
		output->push_back(0);
		++sections[1].pad;
	}

	// insert data
	sections[1].start = sections[0].end;
	sections[1].end = sections[1].start + big_endian(header->dataSize);
	output->insert(output->end(), text + big_endian(header->textSize), text + big_endian(header->textSize) + big_endian(header->dataSize));

	// pad size
	sections[2].pad = sections[1].pad;
	while ( (output->size() & 7) != 0 )
	{
		output->push_back(0);
		++sections[2].pad;
	}

	// insert bss
	sections[2].start = sections[1].end;
	sections[2].end = sections[2].start + big_endian(header->bssSize);

	if (offset)
	{
		uint32_t current = big_endian(*(uint32_t*)&text[offset]);
		uint32_t *ptr = (uint32_t *)(output->data() + shifted_address(offset, sections));
		*ptr = big_endian(org + shifted_address(current, sections));

		while ( *relocTable )
		{
			uint8_t lFix = *relocTable++;
			if( lFix == 1 )
				offset += 254;
			else
			{
				offset += lFix;
				if( offset < max )
				{
					current = big_endian(*(uint32_t*)&text[offset]);
					ptr = (uint32_t *)(output->data() + shifted_address(offset, sections));
					*ptr = big_endian(org + shifted_address(current, sections));
				}
			}
		}
	}
	
	// nuevo tamaño sin la tabla de simbolos ni la tabla de relocalizacion
	return (uint32_t)output->size();
}

char type(uint16_t type)
{
	char t = '?';
	switch ( type )
	{
		case 0x0100:	// In the BSS
			t = 'B';
			break;
		case 0x0200:	// In program text
			t = 'T';
			break;
		case 0x0280:	// Start of an object module
			t = 'M';
			break;
		case 0x02c0:	// Start of a library
			t = 'L';
			break;
		case 0x0400:	// In DATA segment
			t = 'D';
			break;
		case 0x0800:	// External
			t = 'E';
			break;
		case 0x1000:	// Register
			t = 'R';
			break;
		case 0x2000:	// Globales Symbol
			t = 'G';
			break;
		case 0x4000:	// Equated
			t = 'E';
			break;
		case 0x8000:	// Defined
			t = 'd';
			break;
	}

	return t;
}

void show_symbols(void *buffer, uint32_t address)
{
	atari_header *header = (atari_header *)buffer;
	
	if ( header->magic != big_endian((uint16_t)0x601a) )
	{
		std::cout << "Invalid TOS header" << std::endl;
		return;
	}
	
	uint8_t *text = (uint8_t *)buffer + sizeof(atari_header);
	atari_symbol *first = (atari_symbol *)(text + big_endian(header->textSize) + big_endian(header->dataSize));
	atari_symbol *last = (atari_symbol *)(text + big_endian(header->textSize) + big_endian(header->dataSize) + big_endian(header->symbolTableSize));
	
	char name[10];
	memset(name, 0, sizeof(name));
	while ( first != last )
	{
		strncpy(name, &first->name[0], 8);
		printf("%s %c %08x -> %08x\n", name, type(big_endian(first->type)), big_endian(first->value), address + big_endian(first->value));
		
		++first;
	}
}

int main(int argc, const char * argv[])
{
	if ( argc != 3 )
	{
		// insert code here...
		std::cout << "tos2jag input output" << std::endl;
		return -1;
	}

	std::ifstream in;
	in.open(argv[1], std::ios_base::binary);
	if ( !in.is_open() )
	{
		std::cout << "Can't open input file: " << argv[1] << std::endl;
		return -1;
	}

	in.seekg(0, std::ios_base::end);
	auto len = in.tellg();
	char *buffer = new char[len];
	in.seekg(0, std::ios_base::beg);
	in.read(buffer, len);
	in.close();
	
	uint32_t org = 0x4000;
	std::vector<uint8_t> output;
//	show_symbols(buffer, address);
	uint32_t size = relocate_jaguar(buffer, org, &output);
	
	if ( size )
	{
		std::ofstream out;
		out.open(argv[2], std::ios_base::binary | std::ios_base::trunc);
		if ( !out.is_open() )
		{
			std::cout << "Can't open output file: " << argv[2] << std::endl;
			return -1;
		}

		out.write((const char *)output.data(), size);
	}
	
	delete[] buffer;
	return 0;
}

